from abc import ABCMeta, abstractmethod
from threading import Lock
from time import time

from .rate import LimitRate


class Storage(metaclass=ABCMeta):
    @abstractmethod
    def set(self, key, name, value):
        pass

    @abstractmethod
    def get(self, key, name):
        pass


class MemoryStorage(Storage):
    def __init__(self):
        self.__data = {}

    def set(self, key, name, value):
        if key not in self.__data:
            self.__data[key] = {}
        self.__data[key][name] = value

    def get(self, key, name):
        return self.__data.get(key, {}).get(name)


class Limiter(metaclass=ABCMeta):
    @abstractmethod
    def get_current_token(self):
        pass

    @abstractmethod
    def acquire(self, key: str, tokens: int):
        pass


def synchronized(func):
    def _wrapper(*args, **kwargs):
        with Lock():
            return func(*args, **kwargs)
    return _wrapper


class LeakyBucketLimiter(Limiter):
    """漏桶限流器"""
    def __init__(self, capacity: int, rate: LimitRate):
        self.__capacity = capacity
        self.__rate = rate
        self.__last_time = 0
        self.__current_tokens = 0

    @property
    def capacity(self):
        return self.__capacity

    @property
    def rate(self):
        return self.__rate

    @property
    def last_time(self):
        return self.__last_time

    @property
    def current_tokens(self):
        return self.__current_tokens

    @staticmethod
    def timestamp():
        """时间戳:单位毫秒"""
        return int(time() * 1000)

    def set_last_time(self, last_time):
        self.__last_time = last_time

    def set_current_tokens(self, current_tokens):
        self.__current_tokens = current_tokens

    def get_current_token(self):
        timestamp = self.timestamp()
        delta = self.rate * (timestamp - self.last_time)
        self.set_current_tokens(max(0, self.current_tokens - delta))
        self.set_last_time(timestamp)
        return self.current_tokens

    @synchronized
    def acquire(self, key: str, tokens: int):
        current_tokens = self.get_current_token()
        if current_tokens + tokens <= self.capacity:
            self.set_current_tokens(current_tokens + tokens)
            return True
        else:
            return False


class TokenBucketLimiter(Limiter):
    """令牌桶限流器"""

    def __init__(self, capacity: int, rate: LimitRate):
        self.__capacity = capacity
        self.__rate = rate
        self.__last_time = 0
        self.__current_tokens = 0

    @property
    def capacity(self):
        return self.__capacity

    @property
    def rate(self):
        return self.__rate

    @property
    def last_time(self):
        return self.__last_time

    @property
    def current_tokens(self):
        return self.__current_tokens

    @staticmethod
    def timestamp():
        """时间戳:单位毫秒"""
        return int(time() * 1000)

    def set_last_time(self, last_time):
        self.__last_time = last_time

    def set_current_tokens(self, current_tokens):
        self.__current_tokens = current_tokens

    def get_current_token(self):
        timestamp = self.timestamp()
        delta = self.rate * (timestamp - self.last_time)
        self.set_current_tokens(min(self.current_tokens + delta, self.capacity))
        self.set_last_time(timestamp)
        return self.current_tokens

    @synchronized
    def acquire(self, key: str, tokens: int):
        current_tokens = self.get_current_token()
        if  tokens <= current_tokens:
            self.set_current_tokens(current_tokens - tokens)
            return True
        else:
            return False



from open_box.redis import get_redis_client
from open_box.redis.cache import WrapperCache

startup_nodes = [
    {'host': '<host1>', 'port': '<port1>'},
    {'host': '<host2>', 'port': '<port2>'},
    {'host': '<host3>', 'port': '<port3>'},
]
rc = get_redis_client(startup_nodes)
cache = WrapperCache(rc)


def cache_context_usage(**kwargs):
    with cache.CacheContext(key_prefix='good', timeout=5, **kwargs) as ctx:
        if ctx.val:
            print('[from cache]{}: {}'.format(ctx.key, ctx.val))
        else:
            ctx.val = 100
            print('[to cache]{}: {}'.format(ctx.key, ctx.val))


@cache.CacheDecorator(key_prefix='hnf', timeout=5)
def cache_decorator_usage(a, b=3, *args, **kwargs):
    return {'a': a, 'b': b, 'args': args, 'kwargs': kwargs}
    # return 1
    # return datetime.now()


if __name__ == '__main__':
    cache_context_usage(name='hnf', age=28)
    print(cache_decorator_usage('lx', 299, age=28))


from io import BytesIO

import xlsxwriter


def get_format(workbook, bold=False):
    f = workbook.add_format()  # 样式
    if bold:
        f.set_bold()
    f.set_font_name('Arial')  # 字体名称
    f.set_font_size(11)  # 字体大小
    f.set_align('left')  # 水平置中
    f.set_align('vcenter')  # 垂直置中
    f.set_font_color('black')  # 字体颜色
    return f


def build_sheet(workbook, sheet_name, header, data, header_format, body_format):
    worksheet = workbook.add_worksheet(sheet_name)
    worksheet.write_row('A1', header, header_format)
    for i, item in enumerate(data):
        worksheet.write_row('A{0}'.format(i + 2), item, body_format)


def build_xlsx(xlsx_filename_or_fp, sheet_info_list):
    workbook = xlsxwriter.Workbook(xlsx_filename_or_fp)
    header_format = get_format(workbook, bold=True)
    body_format = get_format(workbook, bold=False)

    for b in sheet_info_list:
        sheet_name = b.get('sheet_name')
        header = b.get('header')
        data = b.get('data')
        build_sheet(workbook, sheet_name, header, data, header_format, body_format)

    workbook.close()


def xlsx_for_file(build_data):
    file = BytesIO()
    build_xlsx(file, build_data)
    file.seek(0)
    return file



from open_box.datetime import *

dt = now()
ts = timestamp()
print('now: ', dt)
print('today: ', today())
print('yesterday: ', yesterday())
print('timestamp: ', ts)
print('from_timestamp: ', from_timestamp(ts))
print('to_timestamp: ', to_timestamp(dt))
print('format_datetime: ', format_datetime(dt))
print('to_datetime: ', to_datetime('1992-03-14 21:45:10'))
print(
    'covert_message_duration_text: ',
    covert_message_duration_text(dt - timedelta(seconds=20)),
    covert_message_duration_text(dt - timedelta(minutes=28, seconds=10)),
    covert_message_duration_text(dt - timedelta(hours=7, minutes=28, seconds=10)),
    covert_message_duration_text(dt - timedelta(days=1, hours=7, minutes=28, seconds=10)),
    covert_message_duration_text(dt - timedelta(days=2, hours=7, minutes=28, seconds=10)),
)
print(
    'covert_duration_text: ',
    covert_duration_text(4235, with_unit=False),
    covert_duration_text(4235),
    covert_duration_text(89023, with_unit=False),
    covert_duration_text(89023),
)




from open_box.xlsx.reader import gen_excel_row


def print_every_row(excel_path):
    for row in gen_excel_row(excel_path):
        print(row)


if __name__ == '__main__':
    excel_path = '用户报表.xlsx'
    print_every_row(excel_path)


import zipfile
from io import BytesIO
from logging import INFO
from urllib.request import urlopen

from open_box.log import Logger
from open_box.log.formatter import DEFAULT_TEXT_FORMAT

zip_logger = Logger('zip', level=INFO, formatter=DEFAULT_TEXT_FORMAT)


class Zip(object):
    def __init__(self):
        self.__io = BytesIO()
        self.__zf = zipfile.ZipFile(self.__io, 'a', zipfile.ZIP_DEFLATED, False)

    def add_file(self, alias, filename_or_fp):
        """
        add file to zip
        :param alias: alias name
        :param filename_or_fp: local file, url, fp
        :return:
        """
        if alias not in self.__zf.namelist():
            if isinstance(filename_or_fp, BytesIO):
                self.__add_fp(alias, filename_or_fp)
            elif isinstance(filename_or_fp, str):
                if filename_or_fp.startswith(('http', 'https')):
                    self.__add_url(alias, filename_or_fp)
                else:
                    self.__add_local_file(alias, filename_or_fp)
            else:
                raise Exception('filename_or_fp must be BytesIO or filename')
            zip_logger.info('add [{}] to zip'.format(alias))
        return self

    def __add_fp(self, alias, fp):
        self.__zf.writestr(zipfile.ZipInfo(alias), fp.getvalue(), zipfile.ZIP_DEFLATED)
        fp.seek(0)

    def __add_local_file(self, alias, filename):
        self.__zf.write(filename, alias, zipfile.ZIP_DEFLATED)

    def __add_url(self, alias, url):
        r = urlopen(url)
        if r.status == 200:
            bio = BytesIO()
            bio.write(r.read())
            bio.seek(0)
            self.__add_fp(alias, bio)
        else:
            raise Exception('{} is not a valid url')

    def __close(self):
        self.__zf.close()
        self.__io.seek(0)

    def output(self):
        self.__close()
        return self.__io

    @property
    def size(self):
        self.__io.seek(0, 2)
        size = self.__io.tell()
        self.__io.seek(0)
        return size



from tempfile import NamedTemporaryFile

from open_box.conf.yaml_loader import YamlConfigLoader


def test1():
    with NamedTemporaryFile(suffix='.yaml') as f:
        s = b"""mongodb:
  database_url: mongodb://xxxxxxxxxxxxx
  database_name: yyyyyyyyyyy
elasticsearch:
  hosts: ["127.0.0.1:9200", "127.0.0.1:9201"]
mysql:
  host: mysql+pymysql://zzzzzzzzzzzz
"""
        f.write(s)
        f.seek(0)

        conf = YamlConfigLoader(f.name)

        # 加载完成后直接根据yaml文件的定义来使用配置项
        print(conf.mongodb.database_url)
        print(conf.mongodb.database_name)
        print(conf.elasticsearch.hosts)
        print(conf.mysql.host)

        print(dir(conf))


if __name__ == '__main__':
    test1()



import logging
import os
from logging.handlers import RotatingFileHandler, TimedRotatingFileHandler, SMTPHandler

B = 1
KB = 1024 * B
MB = 1024 * KB
GB = 1024 * MB

SECONDS = 'S'
MINUTES = 'M'
HOURS = 'H'
DAYS = 'D'


class Logger(logging.Logger):
    def __init__(self, name, level=logging.NOTSET, formatter=None):
        super().__init__(name, level)
        self.__formatter = formatter
        self.add_handler(logging.StreamHandler())

    @property
    def formatter(self):
        return self.__formatter

    def add_handler(self, hdr, level=logging.NOTSET, formatter=None):
        hdr.setLevel(level or self.level)
        hdr.setFormatter(formatter or self.__formatter)
        self.addHandler(hdr)
        return self

    def with_default_formatter(self):
        for hdr in self.handlers:
            hdr.setFormatter(self.__formatter)

    def add_rotating_file_handler(self, log_path, level=None, max_bytes=10 * MB, backup_count=5, formatter=None):
        dirname = os.path.dirname(log_path)
        if not os.path.exists(dirname):
            os.makedirs(dirname)
        hdr = RotatingFileHandler(log_path, maxBytes=max_bytes, backupCount=backup_count)
        return self.add_handler(hdr, level, formatter)

    def add_timed_rotating_file_handler(self, log_path, level=None, when=DAYS, interval=1, backup_count=5,
                                        formatter=None):
        dirname = os.path.dirname(log_path)
        if not os.path.exists(dirname):
            os.makedirs(dirname)
        hdr = TimedRotatingFileHandler(log_path, when=when, interval=interval, backupCount=backup_count)
        return self.add_handler(hdr, level, formatter)

    def add_email_handler(self, sender, receives, subject, mail_host, username=None, password=None, level=None,
                          formatter=None):
        hdr = SMTPHandler(
            mailhost=mail_host, fromaddr=sender, toaddrs=receives, subject=subject,
            credentials=(username, password),
        )
        return self.add_handler(hdr, level, formatter)



from rediscluster import RedisCluster


def get_redis_client(startup_nodes=None):
    if startup_nodes is None:
        startup_nodes.append({'host': '127.0.0.1', 'port': '6379'})
    return RedisCluster(
        startup_nodes=startup_nodes,
        decode_responses=True,
        socket_connect_timeout=3,
        socket_timeout=3,
    )



from io import BytesIO

from open_box.email import Mail
from open_box.email import Msg


def send_email():
    from_ = 'no-reply@17zuoye.com'
    host = 'smtp.dc.17zuoye.net'

    with BytesIO() as f:
        s = b"""mongodb:
  database_url: mongodb://xxxxxxxxxxxxx
  database_name: yyyyyyyyyyy
elasticsearch:
  hosts: ["127.0.0.1:9200", "127.0.0.1:9201"]
mysql:
  host: mysql+pymysql://zzzzzzzzzzzz
    """
        f.write(s)
        f.seek(0)

        # 制定邮件信息
        msg = Msg()
        msg.set_sender(
            from_,
        ).set_receivers(  # 收件人
            'han1@example.com',
            'han2@example.com',
        ).set_cc(  # 抄送
            'han3@example.com',
            'han4@example.com',
        ).set_bcc(  # 秘密抄送
            'han5@example.com',
        ).set_subject(  # 主题
            '淦',
        ).set_body(  # 内容
            '来，一起淦！！！',
        ).add_attachment(  # 附件
            f,
            'conf.yaml',
        )

        # 发送邮件
        with Mail(host=host) as email:
            email.send(msg)


if __name__ == '__main__':
    send_email()



from time import sleep

from open_box.redis import get_redis_client

startup_nodes = [
    {'host': '<host1>', 'port': '<port1>'},
    {'host': '<host2>', 'port': '<port2>'},
    {'host': '<host3>', 'port': '<port3>'},
]


def redis_basic():
    rc = get_redis_client(startup_nodes)

    print(rc.set('hnf', 100, 3))
    print(rc.get('hnf') == 100)
    print(rc.set('hnf', 20, 3))
    sleep(3)
    print(rc.get('hnf') == 3)


if __name__ == '__main__':
    redis_basic()
    from uuid import uuid4
    a = uuid4()
    print(str(a), type(str(a)))


from datetime import datetime, date, timedelta

DEFAULT_DATETIME_FORMAT_PRECISION = '%Y-%m-%d %H:%M:%S.%f'
DEFAULT_DATETIME_FORMAT = '%Y-%m-%d %H:%M:%S'
DATE_FORMAT = '%Y-%m-%d'
TIME_FORMAT = '%H:%M:%S'


def now():
    return datetime.now()


def timestamp():
    return datetime.now().timestamp()


def today():
    return date.today()


def yesterday():
    return today() - timedelta(days=1)


def from_timestamp(ts):
    return datetime.fromtimestamp(ts)


def to_timestamp(dt):
    return dt.timestamp()


def format_datetime(dt, format_=DEFAULT_DATETIME_FORMAT_PRECISION):
    return dt.strftime(format_)


def to_datetime(dt_str, format_=DEFAULT_DATETIME_FORMAT):
    return datetime.strptime(dt_str, format_)


def cal_days(seconds):
    return int(seconds // 86400), seconds % 86400


def cal_hours(seconds):
    return int(seconds // 3600), seconds % 3600


def cal_minutes(seconds):
    return int(seconds // 60), seconds % 60


def covert_duration_text(seconds, with_unit=True):
    """转换时长文本"""
    hours, seconds = cal_hours(seconds)
    minutes, seconds = cal_minutes(seconds)
    if with_unit:
        text_list = []
        if hours:
            text_list.extend([str(hours), '小时'])
        if minutes:
            text_list.extend([str(minutes), '分钟'])
        if seconds:
            text_list.extend(['%d' % round(seconds), '秒'])
        return ''.join(text_list)

    return '%02d:%02d:%02d' % (hours, minutes, round(seconds))


def covert_message_duration_text(dt):
    """转换消息发送后距当前时间间隔文本"""
    ts = to_timestamp(dt)
    nts = timestamp()
    delta = nts - ts
    if delta < 60:
        return '刚刚'
    elif 60 <= delta < 3600:
        return '{0}分钟前'.format(int(delta // 60))
    elif 3600 <= delta < 86400:
        return '{0}小时前'.format(int(delta // 3600))
    elif 86400 <= delta < 86400 * 2:
        return '昨天 {}'.format(format_datetime(dt, TIME_FORMAT))
    else:
        return format_datetime(dt, DEFAULT_DATETIME_FORMAT)


from tempfile import NamedTemporaryFile

from open_box.file import File


def file_usage():
    # load file from url
    f = File("https://timgsa.baidu.com/timg?image&quality=80&size=b9999_10000&sec=1603282675348&di=517865944433fc76b375dc7f2415650c&imgtype=0&src=http%3A%2F%2Fa2.att.hudong.com%2F36%2F48%2F19300001357258133412489354717.jpg")
    print(f.filename, f.content_type, f.size, f.size_text, f.md5, f.ext)
    # load file from local path
    f = File(__file__)
    print(f.filename, f.content_type, f.size, f.size_text, f.md5, f.ext)
    # load file from bytes or file-like
    with NamedTemporaryFile(suffix='.yaml') as f:
        s = b"""mongodb:
      database_url: mongodb://xxxxxxxxxxxxx
      database_name: yyyyyyyyyyy
    elasticsearch:
      hosts: ["127.0.0.1:9200", "127.0.0.1:9201"]
    mysql:
      host: mysql+pymysql://zzzzzzzzzzzz
    """
        f.write(s)
        f.seek(0)
        fp = File(f)
        print(fp.filename, fp.content_type, fp.size, fp.size_text, fp.md5, fp.ext)
        fp = File(f.read())
        print(fp.filename, fp.content_type, fp.size, fp.size_text, fp.md5, fp.ext)
        print(fp.profile)


if __name__ == '__main__':
    file_usage()


import pyaml

from . import ConfigObject


class YamlConfigLoader:
    def __init__(self, conf_file):
        with open(conf_file) as f:
            yaml = pyaml.yaml.safe_load(f) or {}
        self.__dict__.update(ConfigObject(**yaml))



from contextlib import contextmanager

from bson import ObjectId
from rediscluster import RedisCluster


class Locker:
    LOCKER_RELEASE_LOCAL_LUA_SCRIPT = """
        if redis.call('get', KEYS[1]) == ARGV[1]
            then
                return redis.call('del', KEYS[1])
            else
                return 0
        end
    """

    def __init__(self, rc: RedisCluster):
        self.__rc = rc

    @contextmanager
    def lock(self, key, timeout=10):
        """
        :param key: lock_name
        :param timeout: timeout
        :return:
        """

        locker_key = 'LOCKER::{0}'.format(key)
        value = str(ObjectId())
        try:
            yield self.__rc.set(locker_key, value, nx=True, ex=timeout)
        finally:
            # 设置val的值，防止任务超时后执行完任务，finally删除当前正在被锁住的任务的key
            # 使用lua脚本，原子化get and del 操作，避免任务执行完删除别人任务的锁的情况
            self.__rc.eval(self.LOCKER_RELEASE_LOCAL_LUA_SCRIPT, 1, locker_key, value)



import inspect
import json
from functools import wraps, partial
from logging import INFO

from rediscluster import RedisCluster

from open_box.log import Logger
from open_box.log.formatter import DEFAULT_TEXT_FORMAT

cache_logger = Logger('cache', level=INFO, formatter=DEFAULT_TEXT_FORMAT)


def check_params(params):
    if not isinstance(params, (int, float, list, dict, str, tuple, bool, type(None))):
        cache_logger.warning('build key failed: {}'.format('{}({}), is not JSON serializable'.format(params, type(params))))
        return False

    if isinstance(params, dict):
        for k, v in params.items():
            code = check_params(v)
            if not code:
                return False

    if isinstance(params, list):
        for param in params:
            code = check_params(param)
            if not code:
                return False
    return True


def json_result(result):
    if not result:
        return result
    try:
        return json.dumps(result, ensure_ascii=False)
    except Exception as ex:
        cache_logger.warning('cache result failed: {}'.format(ex))
        return None


def build_key(key_prefix, params, func_name=None):
    kv_pairs = []
    for k, v in sorted(params.items(), key=lambda a: a[0]):
        val = v
        if isinstance(v, (list, tuple)):
            val = ','.join([str(i) for i in v])
        kv_pairs.append('{}={}'.format(k, val))
    if func_name:
        key = '{}::{}::{}'.format(key_prefix, func_name, '&'.join(kv_pairs))
    else:
        key = '{}::{}'.format(key_prefix, '&'.join(kv_pairs))
    return key


class CacheContext:
    def __init__(self, rc: RedisCluster, key_prefix: str, timeout: int = 600, **params):
        self.__rc = rc
        self.__params = params
        self.__key_prefix = key_prefix
        self.__key = None
        self.val = None
        self.__timeout = timeout
        self.__exist_val = False

    @property
    def key(self):
        return self.__key

    @property
    def key_prefix(self):
        return self.__key_prefix

    def __enter__(self):
        code = check_params(self.__params)
        if code:
            self.__key = build_key(self.__key_prefix, self.__params)
            result_str = self.__rc.get(self.__key)
            if result_str is not None:
                self.val = json.loads(result_str)
                self.__exist_val = True
                cache_logger.info("GET: 【{}】【{}】".format(self.__key, result_str))
            else:
                self.val = None
                self.__exist_val = False
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_val:
            cache_logger.error('cache value failed: {}'.format(exc_val))
            return True
        if self.__key and self.val is not None and not self.__exist_val:
            result_str = json_result(self.val)
            if result_str:
                self.__rc.set(self.__key, result_str, self.__timeout)
                cache_logger.info("SET: 【{}】【{}】".format(self.__key, result_str))
        return True


def load_params(func, *args, **kwargs):
    var_positional_key = None
    positional_or_keyword_keys = []
    keyword_vals = {}
    func_sign = inspect.signature(func)
    for k, v in func_sign.parameters.items():
        if v.kind == inspect._ParameterKind.VAR_POSITIONAL:
            var_positional_key = k

        if v.kind == inspect._ParameterKind.KEYWORD_ONLY:
            keyword_vals[k] = v

        if v.kind == inspect._ParameterKind.POSITIONAL_OR_KEYWORD:
            if v.default == inspect._empty:
                positional_or_keyword_keys.append(k)
            else:
                keyword_vals[k] = v.default

    params = {}
    # positional_or_keyword_params
    positional_or_keyword_params = dict(zip(positional_or_keyword_keys, args))
    params.update(positional_or_keyword_params)

    # var_keyword_params
    var_keyword_params = keyword_vals
    params.update(var_keyword_params)

    # var_positional_params
    var_positional_val = list(args[len(positional_or_keyword_params):])
    if var_positional_key and var_positional_val:
        var_positional_params = {var_positional_key: var_positional_val}
        params.update(var_positional_params)

    # kwargs
    params.update(kwargs)
    return params


def cache_decorator(rc: RedisCluster, key_prefix: str, timeout: int = 600):
    def wrapper(func):
        @wraps(func)
        def __wrapper(*args, **kwargs):
            params = load_params(func, *args, **kwargs)
            code = check_params(params)
            if code:
                key = build_key(key_prefix, params, func.__name__)
                result_str = rc.get(key)
                if result_str is not None:
                    result = json.loads(result_str)
                    cache_logger.info("GET: 【{}】【{}】".format(key, result_str))
                    return result

            result = func(*args, **kwargs)
            if code:
                result_str = json_result(result)
                if result_str:
                    rc.set(key, result_str, timeout)
                    cache_logger.info("SET: 【{}】【{}】".format(key, result_str))
            return result
        return __wrapper
    return wrapper


class WrapperCache:
    def __init__(self, rc: RedisCluster):
        self.CacheContext = partial(CacheContext, rc)
        self.CacheDecorator = partial(cache_decorator, rc)



from functools import wraps
from time import sleep

from ..log import Logger

retry_logger = Logger('retry')


class OverRetryError(Exception):
    def __init__(self, msg=''):
        self.message = 'OverRetryException: {}'.format(msg)

    def __str__(self):
        return self.message


class RetryError(Exception):
    def __init__(self, msg=''):
        self.message = 'RetryError: {}'.format(msg)

    def __str__(self):
        return self.message


def retry(attempts: int = 3, interval: int = 0):
    """
    重试装饰器

    :param attempts: 重试次数
    :param interval: 重试间隔
    :return:
    """

    def wrapper(func):
        @wraps(func)
        def deco(*args, **kwargs):
            last_ex = None
            for i in range(attempts if attempts > 1 else 1):
                try:
                    return func(*args, **kwargs)
                except Exception as ex:
                    last_ex = ex
                    retry_logger.info('{} attempt failed. error: {}'.format(i + 1, ex))
                    if interval:
                        sleep(interval)
            raise OverRetryError(
                'Over {} attempts to retry {}. Error: {}'.format(
                    attempts,
                    func.__name__,
                    last_ex,
                )
            )

        return deco

    return wrapper



from concurrent.futures.thread import ThreadPoolExecutor
from functools import partial
from time import sleep

from open_box.redis import get_redis_client
from open_box.redis.locker import Locker

startup_nodes = [
    {'host': '<host1>', 'port': '<port1>'},
    {'host': '<host2>', 'port': '<port2>'},
    {'host': '<host3>', 'port': '<port3>'},
]
rc = get_redis_client(startup_nodes)
locker = Locker(rc)


def func(locker_name, i):
    with locker.lock(locker_name) as loc:
        if loc:
            sleep(1)
            print('[{}]handle {}'.format(i, locker_name))
        else:
            print('[{}]miss {}'.format(i, locker_name))
            sleep(1)
            func(locker_name, i)


def test_locker():
    with ThreadPoolExecutor(max_workers=3) as exe:
        exe.map(partial(func, 'job'), [1, 2, 3])


if __name__ == '__main__':
    test_locker()



from io import BytesIO
from tempfile import NamedTemporaryFile

from open_box.zip import Zip


def package_file():
    z = Zip()
    with NamedTemporaryFile(prefix='test-', suffix='zip.txt') as f:
        with open(f.name, 'w') as fi:
            fi.write('我是韩能放')
        z.add_file('本地文件.txt', f.name)

    bio = BytesIO()
    bio.write('韩能放'.encode())
    z.add_file('fp.txt', bio)

    z.add_file('远程文件.jpg', 'https://www.baidu.com/img/PCtm_d9c8750bed0b3c7d089fa7d55720d6cf.png')

    output_io = z.output()

    with open('test.zip', 'wb') as f:
        f.write(output_io.getvalue())


if __name__ == '__main__':
    package_file()



ERR_FILE_INPUT_TYPE = 100
ERR_GET_FILE_FROM_URL = 101
ERR_UNKNOWN = -1

ERRORS = {
    ERR_FILE_INPUT_TYPE: '文件初始化输入数据类型错误',
    ERR_GET_FILE_FROM_URL: '从url获取文件失败',
    ERR_UNKNOWN: '未知错误',
}


class ImageError(Exception):
    def __init__(self, msg, code=None, extras=None):
        """
        :param msg: 具体错误信息
        :param code: 错误码
        :param extras:
        """
        self.msg = msg
        if code not in ERRORS:
            code = ERR_UNKNOWN

        self.code = code
        self.code_error = ERRORS.get(self.code)
        self.extras = extras
        error_msg = '[{0}][{1}]{2}'.format(self.code, self.code_error, self.msg)
        super().__init__(error_msg)


class FileInputTypeError(ImageError):
    def __init__(self, msg):
        code = ERR_FILE_INPUT_TYPE
        super().__init__(msg, code)


class GetFileFromUrlError(ImageError):
    def __init__(self, msg):
        code = ERR_GET_FILE_FROM_URL
        super().__init__(msg, code)



from tempfile import NamedTemporaryFile

from pymediainfo import MediaInfo

from . import File
from .error import FileInputTypeError
from ..datetime import covert_duration_text


class Audio(File):
    def __init__(self, input, filename=''):
        super().__init__(input, filename)
        if not self.content_type.startswith('audio'):
            raise FileInputTypeError('Audio仅仅支持音频文件输入')
        duration = self.__get_audio_info()
        self.duration = duration

    def __get_audio_info(self):
        duration = 0

        with NamedTemporaryFile(suffix=self.ext) as f:
            with open(f.name, 'wb') as fi:
                fi.write(self.bio.getvalue())

            self.bio.seek(0)

            media_info = MediaInfo.parse(f.name)
            data = media_info.to_data()
            tracks = data['tracks']
            tracks_audio_list = list(filter(lambda a: a['track_type'].lower() == 'audio', tracks))
            if tracks_audio_list:
                tracks_audio = tracks_audio_list[0]
                duration = tracks_audio['duration']
            return duration

    @property
    def duration_text(self):
        return covert_duration_text(self.duration / 1000)

    @property
    def profile(self):
        pf = super().profile
        pf.update(duration=self.duration, duration_text=self.duration_text)
        return pf


import json
import time
from datetime import datetime
from logging import Formatter


def json_defaulter(d):
    if isinstance(d, datetime):
        return d.strftime('%Y-%m-%d %H:%M:%S')
    else:
        return str(d)


class JsonStyle:
    default_format = '{"message": "{message}"}'
    asctime_search = 'asctime'

    def __init__(self, fmt):
        self.keys = fmt.split('|') or []
        self._fmt = self.__build_fmt() or self.default_format

    def usesTime(self):
        return self._fmt.find(self.asctime_search) >= 0

    def format(self, record):
        return self._fmt.format(**record.__dict__)

    def add_key(self, key):
        if key not in self.keys:
            self.keys.append(key)
            self._fmt = self.__build_fmt() or self.default_format

    def __build_fmt(self):
        _fmt = {}
        for key in self.keys:
            key = key.strip()
            _fmt[key] = '{{{}}}'.format(key)
        return '{' + json.dumps(_fmt, ensure_ascii=False, default=json_defaulter) + '}'


class JsonFormatter(Formatter):
    converter = time.localtime

    def __init__(self, fmt, datefmt):
        super().__init__(fmt, datefmt)
        self._style = JsonStyle(fmt)
        self._fmt = self._style._fmt
        self.datefmt = datefmt

    def add_key(self, key):
        self._style.add_key(key)
        return self

    def add_keys(self, *keys):
        for key in keys:
            self.add_key(key)
        return self


DEFAULT_TEXT_LOG_FORMAT = '%(name)s | %(asctime)s | %(levelname)-8s | %(message)s'
DEFAULT_TEXT_DATE_FORMAT = '%d/%m/%y %H:%M:%S'
DEFAULT_TEXT_FORMAT = Formatter(DEFAULT_TEXT_LOG_FORMAT, DEFAULT_TEXT_DATE_FORMAT)
DEFAULT_JSON_LOG_FORMAT = 'name|asctime|levelname|message'
DEFAULT_JSON_DATE_FORMAT = '%Y-%m-%d %H:%M:%S'
DEFAULT_JSON_FORMAT = JsonFormatter(DEFAULT_JSON_LOG_FORMAT, DEFAULT_JSON_DATE_FORMAT)


from PIL import Image as Im

from . import File


class Image(File):
    def __init__(self, input, filename=''):
        super().__init__(input, filename)
        self.im = Im.open(self.bio)
        self.bio.seek(0)
        self.width, self.height = self.im.size
        self.format = self.im.format

    @property
    def profile(self):
        pf = super().profile
        pf.update(width=self.width, height=self.height, format=self.format)
        return pf



from open_box.file.image import Image


def image_usage():
    img = Image('https://timgsa.baidu.com/timg?image&quality=80&size=b9999_10000&sec=1603282675348&di=517865944433fc76b375dc7f2415650c&imgtype=0&src=http%3A%2F%2Fa2.att.hudong.com%2F36%2F48%2F19300001357258133412489354717.jpg')
    print(img.profile)


if __name__ == '__main__':
    image_usage()



import xlrd


def gen_excel_row(excel_path, sheet_names=None):
    if sheet_names is None:
        sheet_names = []

    for sheet in xlrd.open_workbook(excel_path).sheets():
        if sheet_names:
            if sheet.name not in sheet_names:
                continue

        for row in range(sheet.nrows):
            yield sheet.row_values(row)


import json


class ConfigObject(dict):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.load_attributes(kwargs)

    def load_attributes(self, d):
        for k, v in d.items():
            if isinstance(v, dict):
                self[k] = self.__class__(**v)

            elif isinstance(v, list):
                vs = []
                for i in v:
                    if isinstance(i, dict):
                        vs.append(self.__class__(**i))
                    else:
                        vs.append(i)

                self[k] = vs

    def __getattr__(self, name):
        try:
            value = self[name]
            return value
        except Exception as _:
            return None

    def __setattr__(self, key, value):
        self[key] = value

    def __str__(self):
        return json.dumps(self, indent=4, ensure_ascii=False)




from open_box.xlsx.writer import build_xlsx, xlsx_for_file

data = [
    ['hnf', 28, '男'],
    ['haner', 27, '男'],
    ['liangxiao', 28, '女'],
]

build_data = [
    {
        'sheet_name': 'Sheet1',
        'header': [
            '姓名',
            '年龄',
            '性别',
        ],
        'data': data,
    },
]


def output_excel():
    xlsx_filename = '用户报表.xlsx'
    build_xlsx(xlsx_filename, build_data)


def output_excel_fp():
    xlsx_filename = '用户报表2.xlsx'
    fp = xlsx_for_file(build_data)
    with open(xlsx_filename, 'wb') as f:
        f.write(fp.read())


if __name__ == '__main__':
    output_excel()
    output_excel_fp()



import smtplib
from email.header import Header
from email.mime.application import MIMEApplication
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText


class Msg:
    BASE_MSG_OPTIONS = ('From', 'To', 'Subject')
    RECEIVERS_OPTIONS = ('To', 'Cc', 'Bcc')

    def __init__(self):
        self.__msg = MIMEMultipart()

    @property
    def content(self):
        return self.__msg.as_string()

    @property
    def from_(self):
        if 'From' not in self.__msg:
            return None
        return self.__msg['From']

    @property
    def all_receivers(self):
        recvs = set()
        for k in self.RECEIVERS_OPTIONS:
            if k in self.__msg:
                val = self.__msg[k]
                for receiver in val.split(';'):
                    recvs.add(receiver)
        return list(recvs)

    def set_sender(self, sender):
        self.__msg['From'] = sender
        return self

    def set_receivers(self, *receivers):
        self.__msg['To'] = ';'.join(set(receivers))
        return self

    def set_subject(self, subject):
        self.__msg['Subject'] = Header(subject, 'utf-8')
        return self

    def set_cc(self, *cc):
        self.__msg['Cc'] = ';'.join(set(cc))
        return self

    def set_bcc(self, *bcc):
        self.__msg['Bcc'] = ';'.join(set(bcc))
        return self

    def set_body(self, body):
        self.__msg.attach(MIMEText(body, 'plain', 'utf-8'))
        return self

    def set_html_body(self, html):
        self.__msg.attach(MIMEText(html, 'html', 'utf-8'))
        return self

    def add_attachment(self, io, attachment_name):
        att = MIMEApplication(io.getvalue())
        att.add_header('Content-Disposition', 'attachment', filename=attachment_name)
        self.__msg.attach(att)
        io.seek(0)
        return self

    def validate(self):
        for k in self.BASE_MSG_OPTIONS:
            if k not in self.__msg:
                return False, f'msg lack [{k}] option'
        return True, None


class Mail:
    def __init__(self, host, usr=None, pwd=None):
        self.__host = host  # 邮件服务
        self.__usr = usr
        self.__pwd = pwd
        self.__smtp = None

    def send(self, msg):
        code, err = msg.validate()
        if not code:
            raise Exception(err)

        self.__smtp.sendmail(
            msg.from_,
            msg.all_receivers,
            msg.content,
        )

    def __enter__(self):
        self.__smtp = smtplib.SMTP()
        self.__smtp.connect(self.__host)
        if self.__usr and self.__pwd:
            self.__smtp.login(self.__usr, self.__pwd)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type == smtplib.SMTPException:
            raise exc_val
        self.__smtp.close()



from time import sleep

from open_box.redis.limiter import LeakyBucketLimiter, TokenBucketLimiter
from open_box.redis.limiter.rate import LimitRate, SECOND


def test_leaky_bucket_limiter():
    limit_rate = LimitRate(times=5, intervals=SECOND)
    print(limit_rate)
    leaky_bucket_limiter = LeakyBucketLimiter(5, limit_rate)
    key = 'han.api'

    for i in range(100):
        sleep(0.1)
        print(i, leaky_bucket_limiter.acquire(key, 1), leaky_bucket_limiter.current_tokens)


def test_token_bucket_limiter():
    limit_rate = LimitRate(times=10, intervals=SECOND * 86400)
    print(limit_rate)
    token_bucket_limiter = TokenBucketLimiter(5, limit_rate)
    key = 'han.api'

    for i in range(100):
        sleep(0.1)
        print(i, token_bucket_limiter.acquire(key, 1), token_bucket_limiter.current_tokens)


if __name__ == '__main__':
    # test_leaky_bucket_limiter()
    test_token_bucket_limiter()


from logging import INFO

from open_box.log import Logger, SECONDS
from open_box.log.formatter import DEFAULT_TEXT_FORMAT, DEFAULT_JSON_FORMAT


def base():
    logger = Logger('HNF', level=INFO)
    logger.info('hello world!')


def text_formatter():
    logger = Logger('HNF', level=INFO, formatter=DEFAULT_TEXT_FORMAT)
    logger.info('hello world!')


def json_formatter():
    logger = Logger('HNF', level=INFO, formatter=DEFAULT_JSON_FORMAT)
    logger.info('hello world!')


def json_formatter_with_fields():
    json_formatter = DEFAULT_JSON_FORMAT
    json_formatter.add_key('username').add_key('password')
    json_formatter.add_keys('age', 'gender')
    logger = Logger('HNF', level=INFO, formatter=json_formatter)
    logger.info('login succeed!', extra={
        'username': '123', 'password': '456',
        'age': 28, 'gender': 'male',
    })


def with_file_log():
    json_formatter = DEFAULT_JSON_FORMAT
    json_formatter.add_key('username').add_key('password')
    json_formatter.add_keys('age', 'gender')
    logger = Logger('HNF', level=INFO, formatter=json_formatter)
    logger.add_rotating_file_handler('./logs/hnf.log', formatter=DEFAULT_TEXT_FORMAT)
    logger.info('login succeed!', extra={
        'username': '123', 'password': '456',
        'age': 28, 'gender': 'male',
    })


def with_tiemd_file_log():
    json_formatter = DEFAULT_JSON_FORMAT
    json_formatter.add_key('username').add_key('password')
    json_formatter.add_keys('age', 'gender')
    logger = Logger('HNF', level=INFO, formatter=json_formatter)
    logger.add_timed_rotating_file_handler('./logs/hnf.log', when=SECONDS, formatter=DEFAULT_TEXT_FORMAT)
    logger.info('login succeed!', extra={
        'username': '123', 'password': '456',
        'age': 28, 'gender': 'male',
    })


def with_email_log():
    logger = Logger('HNF', level=INFO)

    sender = '<sender>'
    receives = [
        '369685930@qq.com'
    ]
    mail_host = '<mail-host>'
    subject = 'login failed'
    logger.add_email_handler(
        sender, receives, subject, mail_host=mail_host, level=INFO
    )

    logger.error('login failed!', extra={
        'username': '123', 'password': '456',
        'age': 28, 'gender': 'male',
    })


if __name__ == '__main__':
    base()
    # text_formatter()
    # json_formatter()
    # json_formatter_with_fields()
    # with_file_log()
    # with_tiemd_file_log()
    # with_email_log()


import hashlib
import mimetypes
import os
import re
from io import BytesIO
from urllib.parse import unquote
from urllib.request import urlopen

import magic

from .error import GetFileFromUrlError, FileInputTypeError

SIZE_B = 'B'
SIZE_KB = 'KB'
SIZE_MB = 'MB'
SIZE_GB = 'GB'
SIZE_TB = 'TB'
SIZE_PB = 'PB'
UNIT_MAP = {
    0: SIZE_B,
    1: SIZE_KB,
    2: SIZE_MB,
    3: SIZE_GB,
    4: SIZE_TB,
    5: SIZE_PB,
}

B = 1
KB = 1024 * B
MB = 1024 * KB
GB = 1024 * MB
TB = 1024 * GB
PB = 1024 * TB


class File:
    FILENAME_PATTERN = re.compile(r'filename\*=UTF-8\'\'')

    def __init__(self, input, filename=""):
        self.bio, self.filename  = self.__load_file(input, filename)
        self.size = self.__load_size()
        self.content_type = self.__load_content_type(filename)
        self.ext = self.__load_ext()
        self.__refine_filename()

    @property
    def size_text(self):
        num = self.size
        n = 0
        while num > 1024:
            num = num / 1024.0
            n += 1
            if n >= 5:
                break
        return '%.2f%s' % (num, UNIT_MAP.get(n))

    @property
    def md5(self):
        md5 = hashlib.md5(self.bio.getvalue()).hexdigest()
        self.bio.seek(0)
        return md5

    def __refine_filename(self):
        if not self.filename:
            if self.md5 and self.ext:
                self.filename = '{}{}'.format(self.md5, self.ext)

    def __load_ext(self):
        _, ext = os.path.splitext(self.filename)
        if not ext:
            ext = mimetypes.guess_extension(self.content_type) or ''
        return ext

    def __load_file(self, input, filename):
        if isinstance(input, BytesIO):
            return input, filename

        if isinstance(input, bytes):
            bio = BytesIO(input)
            bio.seek(0)
            return bio, filename

        if hasattr(input, 'read') and hasattr(input, 'seek'):
            bio = BytesIO(input.read())
            bio.seek(0)
            input.seek(0)
            return bio, filename

        if isinstance(input, str):
            if os.path.exists(input):
                if not filename:
                    filename = os.path.basename(input)

                bio = BytesIO()
                with open(input, 'rb') as f:
                    bio.write(f.read())
                bio.seek(0)
                return bio, filename

            if input.startswith(("http", "https")):
                r = urlopen(input)
                if r.status == 200:
                    if not filename:
                        filename = self.__get_filename_from_url(r)

                    bio = BytesIO()
                    bio.write(r.read())
                    bio.seek(0)
                    return bio, filename
                else:
                    raise GetFileFromUrlError(input)

        raise FileInputTypeError("")

    def __load_size(self):
        self.bio.seek(0, 2)
        size = self.bio.tell()
        self.bio.seek(0)
        return size

    def __load_content_type(self, filename):
        buf = magic.from_buffer(self.bio.read(), mime=True)
        self.bio.seek(0)
        if buf == 'application/octet-stream':
            if filename:
                _, ext = os.path.splitext(filename)
                mimetype = mimetypes.types_map.get(ext, '')
                if mimetype != buf:
                    return mimetype
        return buf

    def __get_filename_from_url(self, result):
        attachment_name = result.info().get_filename() or ''
        content_disposition = result.headers.get('content-disposition', '')
        if content_disposition:
            for item in content_disposition.split(';'):
                item = item.strip()
                if 'filename*=UTF-8' in item:
                    attachment_name = re.sub(self.FILENAME_PATTERN, '', item)
                    attachment_name = unquote(attachment_name)

        attachment_name = unquote(attachment_name)
        if not attachment_name:
            attachment_name = os.path.basename(result.url)
        return attachment_name

    def save(self, dirname=''):
        if dirname and not os.path.exists(dirname):
            os.makedirs(dirname)

        path = self.filename
        if dirname:
            if not dirname.endswith('/'):
                dirname += '/'
            path = dirname + path

        with open(path, 'wb') as f:
            f.write(self.bio.getvalue())

        self.bio.seek(0)

    @property
    def profile(self):
        return {
            'filename': self.filename,
            'content_type': self.content_type,
            'size': self.size,
            'size_text': self.size_text,
            'md5': self.md5,
            'ext': self.ext,
        }



MILLISECOND = 1
SECOND = MILLISECOND * 1000
MINUTE = SECOND *60
HOUR = MINUTE * 60
DAY = HOUR * 24
MONTH = DAY * 30
YEAR = DAY * 365


class LimitRate:
    def __init__(self, times: int, intervals: int):
        self.__times = times
        self.__intervals = intervals
        self.__rate = self.cal_rate(times, intervals)

    @property
    def times(self):
        return self.__times

    @property
    def intervals(self):
        return self.__intervals

    @property
    def rate(self):
        return self.__rate

    def __str__(self):
        return '<LimitRate: {} times/ms>'.format(self.__rate)

    @staticmethod
    def cal_rate(times: int, intervals: int):
        if intervals == 0:
            return times
        return times / intervals

    def __mul__(self, other):
        return self.__rate * other